from __future__ import unicode_literals

import logging
import os
import pprint
from collections import OrderedDict
from typing import Dict, List, Optional

from .api import ConfigConvert, Dictable, EnvMapper, OdooConfigABC
from .config_section import (
    AddonsPathConfigSection,
    DatabaseOdooConfigSection,
    HttpOdooConfigSection,
    LimitOdooConfigSection,
    LoggerSection,
    OtherSection,
    ServerWideModuleConfigSection,
    UpdateInstallSection,
    WorkersOdooConfigSection,
)
from .mapper import (
    CleverCloudCellarCompatibilityMapper,
    CleverCloudPostgresCompatibilityMapper,
    OdooCompatibilityMapper,
    OdooQueueJobMapper,
    OdooRedisSessionMapper,
)

_logger_level = getattr(logging, os.environ.get("NDP_SERVER_LOG_LEVEL", "INFO"), logging.INFO)

_logger = logging.getLogger("install_addons")
_logger.setLevel(_logger_level)

PG_MAX_CONN_MODE_AUTO = "AUTO"
PG_MAX_CONN_MODE_DEFAULT = "DEFAULT"


def is_main_instance(env_vars):
    return env_vars.get("INSTANCE_NUMBER", 0) == 0


class OdooConfig(OdooConfigABC, Dictable):
    def __init__(self, env_vars, odoo_rc=None, main_instance=None):
        super(OdooConfig, self).__init__(main_instance is None and is_main_instance(env_vars))
        self._odoo_version = self.to_int(env_vars.get("ODOO_VERSION"))
        self.odoo_rc = odoo_rc
        env_converter = OdooEnvConverter()
        env_vars = env_converter.map_env_vars(env_vars)
        env_converter.assert_env(env_vars)

        self.addons_config = AddonsPathConfigSection(self, env_vars)
        self.workers_config = WorkersOdooConfigSection(self, env_vars)
        self.limit_config = LimitOdooConfigSection(self, env_vars)
        self.db_config = DatabaseOdooConfigSection(self, env_vars)
        self.http_config = HttpOdooConfigSection(self, env_vars)
        self.wide_module = ServerWideModuleConfigSection(self, env_vars)
        self.other_section = OtherSection(self, env_vars)
        self.logger_section = LoggerSection(self, env_vars)
        self.update_install = UpdateInstallSection(self, env_vars)

    @property
    def odoo_version(self):
        return self._odoo_version

    def to_dict(self):
        result = OrderedDict()
        result["--config"] = self.odoo_rc
        result.update(self.update_install.to_dict())
        result.update(self.addons_config.to_dict())
        result.update(self.workers_config.to_dict())
        result.update(self.limit_config.to_dict())
        result.update(self.http_config.to_dict())
        result.update(self.db_config.to_dict())
        result.update(self.wide_module.to_dict())
        result.update(self.other_section.to_dict())
        result.update(self.logger_section.to_dict())
        return result

    def __repr__(self):
        return pprint.pformat(self.to_dict())

    def to_odoo_args(self):
        # type: () -> List[str]
        result = []
        dict_values = Dictable.clean_none_env_vars(self.to_dict())
        dict_values = Dictable.clean_config_dict(dict_values)
        for key, value in dict_values.items():
            if not value:
                continue

            if value == str(True):
                result.append(key)
            else:
                result.append("%s=%s" % (key, value))
        return result


class OdooEnvConverter(ConfigConvert):
    def __init__(self, mappers=None):
        # type: (Optional[List[EnvMapper]]) -> OdooEnvConverter
        self.mappers = list(mappers or [])
        self.mappers.append(CleverCloudCellarCompatibilityMapper())
        self.mappers.append(CleverCloudPostgresCompatibilityMapper())
        self.mappers.append(OdooCompatibilityMapper())
        self.mappers.append(OdooRedisSessionMapper())
        self.mappers.append(OdooQueueJobMapper())

    def map_env_vars(self, env_vars):
        # type: (Dict[str, str]) -> Dict[str, str]
        result = dict(env_vars)
        for mapper in self.mappers:
            result.update(mapper.map_vars(env_vars))
        for key in dict(result).keys():
            if not result[key]:
                result.pop(key, None)
        return result

    def assert_env(self, env_vars):
        if not self.is_true(env_vars.get("REMOTE_DB", str(True))):
            return
        if (
            not env_vars.get("DB_NAME")
            or not env_vars.get("DB_PORT")
            or not env_vars.get("DB_HOST")
            or not env_vars.get("DB_USER")
            or not env_vars.get("DB_PASSWORD")
        ):
            raise TypeError(
                "Can't start Odoo without a db name "
                "Please add the one of the following environment variable"
                "- DATABASE"
                "- DB_NAME"
                "- POSTGRESQL_ADDON_DB"
            )

    def create_config_dict(self, env_vars):
        env_vars = self.map_env_vars(env_vars)
        config = OdooConfig(env_vars, is_main_instance(env_vars))
        new_options = config.to_dict()
        new_options = Dictable.clean_config_dict(new_options)
        return new_options



import abc
import sys
from collections import OrderedDict
from typing import Any, Dict, List, Optional, Union

if sys.version_info[0] == 2:  # Python 2

    class ABC(object):
        __meta_class__ = abc.ABCMeta
        __slots__ = ()

else:
    ABC = abc.ABC


class Dictable(ABC):
    def to_dict(self):
        # type: () -> Dict[str, Any]
        raise NotImplementedError()

    @staticmethod
    def clean_config_dict(values):
        # type: (Union[Dict[str, Any], Dictable]) -> Dict[str, str]
        new_values = OrderedDict()
        if isinstance(values, Dictable):
            values = values.to_dict()

        for key, value in values.items():
            if isinstance(value, dict):
                value = Dictable.clean_config_dict(value)
            elif isinstance(value, (list, tuple, set)):
                value = ",".join([str(x) for x in value]) or ""
            if value and value is not None and not isinstance(value, dict):
                new_values[key] = str(value)
        return new_values

    @staticmethod
    def clean_none_env_vars(dict_value):
        # type: (Union[Dict[str, Any], Dictable]) -> Dict[str, Any]
        result = OrderedDict()
        if isinstance(dict_value, Dictable):
            dict_value = dict_value.to_dict()

        for key, value in dict_value.items():
            if value is not None:
                result[key] = value
        return result


class ConfigConvert(ABC):
    def is_true(self, any):
        # type: (Union[str, bool, int, None]) -> bool
        if not any or not isinstance(any, (str, bool, int)):
            return False
        return bool(any) and (str(any).isdigit() and bool(int(any))) or (str(any).capitalize() == str(True)) or False

    def to_int(self, any):
        # type: (Union[str, bool, int, None]) -> int
        if not any or not isinstance(any, (str, bool, int)):
            return 0
        if isinstance(any, str):
            return any.isdigit() and int(any) or "." in any and int(float(any)) or 0
        return int(any)


class EnvMapper(ConfigConvert):
    def map_vars(self, env_vars):
        # type: (Dict[str, str]) -> Dict[str, str]
        raise NotImplementedError()

    def get_value(self, env_vars, possible_value, default=None):
        # type: (Dict[str, str], List[str], str) -> Optional[str]
        for value in possible_value:
            res = env_vars.get(value)
            if res:
                return res
        return default


class OdooConfigABC(ConfigConvert, ABC):
    def __init__(self, main_instance=True):
        super(OdooConfigABC, self).__init__()
        self.main_instance = main_instance

    @property
    def odoo_version(self):
        raise NotImplementedError


class OdooConfigSection(ConfigConvert, Dictable, ABC):
    def __init__(self, odoo_config_maker, env_vars):
        # type: (OdooConfigABC, Dict[str, Union[str, bool, int, None]]) -> None
        self.config_maker = odoo_config_maker
        self.enable = True

    def to_dict(self):
        # type: () -> Dict[str, Any]
        if not self.enable:
            return {}
        return self.get_info()



from typing import Dict, Union

from .api import EnvMapper


class OdooCompatibilityMapper(EnvMapper):
    """
    <WORKER_HTTP> is put in <WORKERS> if <WORKERS> not exist
    and
    <WORKERS> is put in <WORKER_HTTP> if <WORKER_HTTP> not exist
    """

    def map_vars(self, env_vars):
        # type: (Dict[str, str]) -> Dict[str, str]
        return {
            "WORKER_HTTP": self.get_value(env_vars, ["WORKER_HTTP", "WORKERS"]),
            "WORKER_CRON": self.get_value(env_vars, ["WORKER_CRON", "WORKERS"]),
            "WORKER_JOB": self.get_value(env_vars, ["WORKER_JOB", "WORKERS"]),
            "HTTP_INTERFACE": self.get_value(env_vars, ["HTTP_INTERFACE", "XMLRPC_INTERFACE"]),
            "HTTP_PORT": self.get_value(env_vars, ["HTTP_PORT", "XMLRPC_PORT"]),
            "HTTP_ENABLE": self.get_value(env_vars, ["HTTP_ENABLE", "XMLRPC_ENABLE"]),
            "LONGPOLLING_PORT": self.get_value(env_vars, ["LONGPOLLING_PORT"]),
            "SERVER_WIDE_MODULES": self.get_value(env_vars, ["SERVER_WIDE_MODULES", "LOAD"]),
        }


class OdooQueueJobMapper(EnvMapper):
    """
    <WORKER_HTTP> is put in <WORKERS> if <WORKERS> not exist
    and
    <WORKERS> is put in <WORKER_HTTP> if <WORKER_HTTP> not exist
    """

    def map_vars(self, env_vars):
        # type: (Dict[str, str]) -> Dict[str, str]
        enable = self.is_true(env_vars.get("ODOO_QUEUE_JOB_ENABLE"))
        return {
            "ODOO_QUEUE_JOB_ENABLE": str(enable),
            "ODOO_QUEUE_JOB_CHANNELS": enable
            and self.get_value(env_vars, ["ODOO_QUEUE_JOB_CHANNELS", "ODOO_CONNECTOR_CHANNELS"])
            or None,
            "ODOO_QUEUE_JOB_SCHEME": enable
            and self.get_value(env_vars, ["ODOO_QUEUE_JOB_SCHEME", "ODOO_CONNECTOR_SCHEME"])
            or None,
            "ODOO_QUEUE_JOB_HOST": enable
            and self.get_value(env_vars, ["ODOO_QUEUE_JOB_HOST", "ODOO_CONNECTOR_HOST"])
            or None,
            "ODOO_QUEUE_JOB_PORT": enable
            and self.get_value(env_vars, ["ODOO_QUEUE_JOB_PORT", "ODOO_CONNECTOR_PORT"])
            or None,
            "ODOO_QUEUE_JOB_HTTP_AUTH_USER": enable
            and self.get_value(env_vars, ["ODOO_QUEUE_JOB_HTTP_AUTH_USER", "ODOO_CONNECTOR_HTTP_AUTH_USER"])
            or None,
            "ODOO_QUEUE_JOB_HTTP_AUTH_PASSWORD": enable
            and self.get_value(env_vars, ["ODOO_QUEUE_JOB_HTTP_AUTH_PASSWORD", "ODOO_CONNECTOR_HTTP_AUTH_PASSWORD"])
            or None,
            "ODOO_QUEUE_JOB_JOBRUNNER_DB_HOST": enable
            and self.get_value(env_vars, ["ODOO_QUEUE_JOB_JOBRUNNER_DB_HOST", "ODOO_CONNECTOR_JOBRUNNER_DB_HOST"])
            or None,
            "ODOO_QUEUE_JOB_JOBRUNNER_DB_PORT": enable
            and self.get_value(env_vars, ["ODOO_QUEUE_JOB_JOBRUNNER_DB_PORT", "ODOO_CONNECTOR_JOBRUNNER_DB_PORT"])
            or None,
        }


class OdooRedisSessionMapper(EnvMapper):
    """
    <WORKER_HTTP> is put in <WORKERS> if <WORKERS> not exist
    and
    <WORKERS> is put in <WORKER_HTTP> if <WORKER_HTTP> not exist
    """

    def map_vars(self, env_vars):
        # type: (Dict[str, str]) -> Dict[str, str]
        enable = self.is_true(
            env_vars.get("REDIS_SESSION_ENABLE", bool(self.get_value(env_vars, ["REDIS_SESSION_HOST", "REDIS_HOST"])))
        )
        return {
            "REDIS_SESSION_ENABLE": str(enable),
            "REDIS_SESSION_URL": enable and self.get_value(env_vars, ["REDIS_SESSION_URL", "REDIS_URL"]) or None,
            "REDIS_SESSION_HOST": enable and self.get_value(env_vars, ["REDIS_SESSION_HOST", "REDIS_HOST"]) or None,
            "REDIS_SESSION_PORT": enable and self.get_value(env_vars, ["REDIS_SESSION_PORT", "REDIS_PORT"]) or None,
            "REDIS_SESSION_DB_INDEX": enable
            and self.get_value(env_vars, ["REDIS_SESSION_DB_INDEX", "REDIS_DB_INDEX"])
            or None,
            "REDIS_SESSION_PASSWORD": enable
            and self.get_value(env_vars, ["REDIS_SESSION_PASSWORD", "REDIS_PASSWORD"])
            or None,
        }


class CleverCloudCellarCompatibilityMapper(EnvMapper):
    """
    Compatibility mapper to convert Environment varaible provided by Cellar from CleverCloud
    https://www.clever-cloud.com/doc/deploy/addon/cellar/

    IS_TRUE(CELLAR_ADDON_HOST) => S3_FILESTORE_ENABLE
    CELLAR_ADDON_HOST => S3_FILESTORE_HOST
    CELLAR_ADDON_KEY_SECRET => S3_FILESTORE_SECRET_KEY
    CELLAR_ADDON_KEY_ID => S3_FILESTORE_ACCESS_KEY
    CELLAR_ADDON_REGION or "fr-par" => S3_FILESTORE_REGION
    """

    def map_vars(self, env_vars):
        # type: (Dict[str, Union[str, bool, int, None]]) -> Dict[str, Union[str, bool, int, None]]
        enable = self.is_true(
            env_vars.get(
                "S3_FILESTORE_ENABLE", bool(self.get_value(env_vars, ["S3_FILESTORE_HOST", "CELLAR_ADDON_HOST"]))
            )
        )
        return {
            "S3_FILESTORE_ENABLE": str(enable),
            "S3_FILESTORE_HOST": enable
            and self.get_value(env_vars, ["S3_FILESTORE_HOST", "CELLAR_ADDON_HOST"])
            or None,
            "S3_FILESTORE_SECRET_KEY": enable
            and self.get_value(env_vars, ["S3_FILESTORE_SECRET_KEY", "CELLAR_ADDON_KEY_SECRET"])
            or None,
            "S3_FILESTORE_ACCESS_KEY": enable
            and self.get_value(env_vars, ["S3_FILESTORE_ACCESS_KEY", "CELLAR_ADDON_KEY_ID"])
            or None,
            # Pas de region fournit par S3 CleverCloud
            "S3_FILESTORE_REGION": enable
            and self.get_value(env_vars, ["S3_FILESTORE_REGION", "CELLAR_ADDON_REGION"], default="fr-par")
            or None,
        }


class CleverCloudPostgresCompatibilityMapper(EnvMapper):
    def map_vars(self, env_vars):
        return {
            "DB_NAME": self.get_value(env_vars, ["DB_NAME", "DATABASE", "POSTGRESQL_ADDON_DB", "POSTGRES_DB"]),
            "DB_HOST": self.get_value(env_vars, ["DB_HOST", "POSTGRESQL_ADDON_DIRECT_HOST", "POSTGRESQL_ADDON_HOST"]),
            "DB_PORT": self.get_value(env_vars, ["DB_PORT", "POSTGRESQL_ADDON_DIRECT_PORT", "POSTGRESQL_ADDON_PORT"]),
            "DB_USER": self.get_value(env_vars, ["DB_USER", "POSTGRESQL_ADDON_USER", "POSTGRES_USER"]),
            "DB_PASSWORD": self.get_value(env_vars, ["DB_PASSWORD", "POSTGRESQL_ADDON_PASSWORD", "POSTGRES_PASSWORD"]),
        }



from __future__ import unicode_literals

import argparse
import logging
import os
import subprocess
import sys

if sys.version_info >= (3,):
    import configparser
else:
    import ConfigParser as configparser

import uuid
from typing import Dict, List, Optional

from addons_installer import AddonsInstaller, AddonsRegistry

from .odoo_config_maker import OdooConfig

_logger = logging.getLogger("launch")
_logger.setLevel(logging.DEBUG)
handler = logging.StreamHandler()
_logger.addHandler(handler)


class Launcher(object):
    def __init__(self, args, odoo_path=None, odoo_rc=None, server_path=None):
        # type: (List[str], Optional[str], Optional[str], Optional[str]) -> Launcher
        parser = self.get_parser()
        ns, other = parser.parse_known_args(args=args)
        odoo_path = ns.odoo_path or odoo_path
        assert odoo_path, "No Odoo path is provided"
        self.odoo_path = os.path.abspath(os.path.expanduser(odoo_path))
        self.other_args = other

        odoo_rc = ns.odoo_rc or odoo_rc
        assert "No Odoo config file path is provided"
        self.odoo_rc = os.path.abspath(odoo_rc)
        ndp_server_path = ns.server_path or server_path
        assert ndp_server_path, "Server path is provided"
        self.ndp_server_path = os.path.abspath(os.path.expanduser(ndp_server_path))

    def get_parser(self):
        # type: () -> argparse.ArgumentParser
        parser = argparse.ArgumentParser()
        parser.add_argument("--odoo-path", dest="odoo_path", help="Path of odoo-bin")
        parser.add_argument("--odoo-rc", "-c", dest="odoo_rc", help="Path of the base config file")
        parser.add_argument("--ndp-server-path", dest="server_path", help="Server Path")
        return parser

    def init_addons(self, env_vars):
        registry = AddonsRegistry()
        addons = registry.parse_env(env_vars)
        for addon in addons:
            AddonsInstaller.install(addon)

    def create_config_file_args(self, env_vars):
        # type: (Dict[str, str]) -> OdooConfig
        config = OdooConfig(env_vars, self.odoo_rc)
        # config.http_config.enable = False

        config.update_install.enable = False
        config.update_install.save_config_file = True
        config.update_install.force_stop_after_init = True
        config.update_install.force_save_config_file = True
        return config

    def _get_default_config_file(self, env_vars):
        if os.path.exists(self.odoo_rc):
            _logger.info("Remove file %s", self.odoo_rc)
            os.remove(self.odoo_rc)
        default_config = self.get_config_parser(env_vars)
        with open(self.odoo_rc, "w") as fp:
            default_config.write(fp)

    def get_config_parser(self, env_vars):
        default_config = configparser.ConfigParser()
        section = "options"
        default_config.add_section(section)
        default_config.set(section, "admin_passwd", env_vars.get("ADMIN_PASSWD", str(uuid.uuid4())))
        default_config.set(section, "csv_internal_sep", env_vars.get("CSV_INTERNAL_SEP", ","))
        default_config.set(
            section, "publisher_warranty_url", env_vars.get("PUBLISHER_WARRANTY_URL", "https://ndp-systemes.fr")
        )
        default_config.set(section, "reportgz", str(False))
        default_config.set(section, "root_path", str(False))
        return default_config

    def launch_config_file(self, env_vars):
        self._get_default_config_file(env_vars)
        config = self.create_config_file_args(env_vars)
        return self._launch_odoo("ndpserver", config)

    def launch_update(self, env_vars):
        # type: (Dict[str, str]) -> subprocess.Popen
        config = self.config_launch_update(env_vars)
        return self._launch_odoo("ndpserver", config)

    def launch(self, env_vars):
        # type: (Dict[str, str]) -> subprocess.Popen
        return self._launch_odoo("ndpserver", None)

    def config_launch_update(self, env_vars):
        # type: (Dict[str, str]) -> OdooConfig
        config = OdooConfig(env_vars, self.odoo_rc)
        config.http_config.enable = False
        config.workers_config.enable = False
        config.db_config.enable = False
        config.addons_config.enable = False
        config.workers_config.enable = False
        config.update_install.enable = True

        config.wide_module.remove_queue_job()
        config.update_install.stop_after_init = True

        return config

    def launch_maintenance_server(self):
        # type: () -> subprocess.Popen
        return self._execute_popen([sys.executable, "-m", "maintenance_server"])

    def normalize_args(self, srv, config):
        # type: (str, Optional[OdooConfig]) -> List[str]
        cmd_args = [sys.executable, os.path.join(self.odoo_path, "odoo-bin")]
        if srv:
            cmd_args.append("--addons-path=%s" % self.ndp_server_path)
            cmd_args.append(srv)
        return cmd_args + (config and config.to_odoo_args() or [])

    def _launch_odoo(self, srv, config):
        # type: (str, Optional[OdooConfig]) -> subprocess.Popen
        return self._execute_popen(self.normalize_args(srv, config))

    def _execute_popen(self, cmd):
        # type: (List[str]) -> subprocess.Popen
        _logger.info("Run -> %s", " ".join([str(s) for s in cmd]))
        return subprocess.Popen(cmd)



import enum
from collections import OrderedDict
from typing import Dict, Union

from addons_installer import addons_installer

from .api import OdooConfigABC, OdooConfigSection


class WorkersOdooConfigSection(OdooConfigSection):
    class UseCase(enum.Enum):
        CLASSIC = "CLASSIC"
        ONLY_HTTP = "ONLY_HTTP"
        ONLY_JOB_RUNNER = "ONLY_JOB_RUNNER"
        ONLY_JOB_WORKER = "ONLY_JOB_WORKER"
        ONLY_CRON = "ONLY_CRON"

    def __init__(self, odoo_config_maker, env_vars):
        # type: (OdooConfigABC, Dict[str, Union[str, bool, int, None]]) -> None
        super(WorkersOdooConfigSection, self).__init__(odoo_config_maker, env_vars)
        self.http = self.to_int(env_vars.get("WORKER_HTTP", 1))
        self.cron = self.to_int(env_vars.get("WORKER_CRON", 2))
        self.job = self.to_int(env_vars.get("WORKER_JOB", 0))
        self.split_use_case = self.is_true(env_vars.get("SPLIT_ODOO_USE_CASE", str(False)))

        use_case_env = env_vars.get("ODOO_USE_CASE")
        if not use_case_env or use_case_env not in (list(WorkersOdooConfigSection.UseCase)):
            self.odoo_use_case = WorkersOdooConfigSection.UseCase.CLASSIC
        else:
            self.odoo_use_case = WorkersOdooConfigSection.UseCase[use_case_env]

        if not odoo_config_maker.main_instance and self.odoo_use_case != WorkersOdooConfigSection.UseCase.ONLY_CRON:
            self.cron = 0

    @property
    def total(self):
        return self.http + self.cron + self.job

    @property
    def worker(self):
        if self.odoo_use_case == WorkersOdooConfigSection.UseCase.ONLY_JOB_WORKER:
            return self.job
        if self.odoo_use_case == WorkersOdooConfigSection.UseCase.ONLY_JOB_RUNNER:
            return 0
        if self.odoo_use_case == WorkersOdooConfigSection.UseCase.ONLY_HTTP:
            return self.http
        if self.odoo_use_case == WorkersOdooConfigSection.UseCase.ONLY_CRON:
            return 0
        return self.http + self.job

    def to_dict(self):
        if not self.enable:
            return {}
        return {
            "--workers": self.worker,
            "--max-cron-threads": self.cron,
        }


class LimitOdooConfigSection(OdooConfigSection):
    def __init__(self, odoo_config_maker, env_vars):
        super(LimitOdooConfigSection, self).__init__(odoo_config_maker, env_vars)
        self.limit_request = int(env_vars.get("LIMIT_REQUEST", 0)) or None
        self.limit_time_cpu = int(env_vars.get("LIMIT_TIME_CPU", 0)) or None
        self.limit_time_real = int(env_vars.get("LIMIT_TIME_REAL", 0)) or None
        self.osv_memory_count_limit = int(env_vars.get("OSV_MEMORY_COUNT_LIMIT", 0)) or None
        self.osv_memory_age_limit = int(env_vars.get("OSV_MEMORY_AGE_LIMIT", 0)) or None
        self.limit_memory_hard = int(env_vars.get("LIMIT_MEMORY_HARD", 0)) or None
        self.limit_memory_soft = int(env_vars.get("LIMIT_MEMORY_SOFT", 0)) or None

        if not self.limit_memory_hard or not self.limit_memory_soft:
            global_limit_memory_hard = int(env_vars.get("GLOBAL_LIMIT_MEMORY_HARD", 0))
            global_limit_memory_soft = int(env_vars.get("GLOBAL_LIMIT_MEMORY_SOFT", 0))
            total = WorkersOdooConfigSection(odoo_config_maker, env_vars).total or 1
            if not self.limit_memory_soft and global_limit_memory_soft:
                self.limit_memory_soft = global_limit_memory_soft // total
            if not self.limit_memory_hard and global_limit_memory_hard:
                self.limit_memory_hard = global_limit_memory_hard // total

    def to_dict(self):
        return {
            "--limit-request": self.limit_request,
            "--limit-time-cpu": self.limit_time_cpu,
            "--limit-time-real": self.limit_time_real,
            "--limit-memory-hard": self.limit_memory_hard,
            "--limit-memory-soft": self.limit_memory_soft,
            "--osv-memory-count-limit": self.osv_memory_count_limit,
            "--osv-memory-age-limit": self.osv_memory_age_limit,
        }


class DatabaseOdooConfigSection(OdooConfigSection):
    class MaxConnMode(enum.Enum):
        AUTO = "AUTO"
        FIXED = "FIXED"

    def __init__(self, odoo_config_maker, env_vars):
        super(DatabaseOdooConfigSection, self).__init__(odoo_config_maker, env_vars)
        self.name = env_vars.get("DB_NAME")
        self.host = env_vars.get("DB_HOST")
        self.port = self.to_int(env_vars.get("DB_PORT")) or None
        self.user = env_vars.get("DB_USER")
        self.password = env_vars.get("DB_PASSWORD")
        self.max_conn = self.to_int(env_vars.get("DB_MAX_CONN"))
        self.filter = env_vars.get("DB_FILTER")
        self.log_enable = env_vars.get("LOG_DB")
        self.log_level = env_vars.get("LOG_DB_LEVEL")
        self.show = self.is_true(env_vars.get("LIST_DB"))
        mode_env = env_vars.get("DB_MAX_CONN_MODE")
        if not mode_env or mode_env not in (
            DatabaseOdooConfigSection.MaxConnMode.FIXED.value,
            DatabaseOdooConfigSection.MaxConnMode.AUTO.value,
        ):
            mode = DatabaseOdooConfigSection.MaxConnMode.AUTO
        else:
            mode = DatabaseOdooConfigSection.MaxConnMode[mode_env]

        nb_workers = WorkersOdooConfigSection(odoo_config_maker, env_vars).total or 1
        min_conn = nb_workers + int(nb_workers // 2)
        if mode == DatabaseOdooConfigSection.MaxConnMode.FIXED and not self.max_conn:
            # Switch to auto if no max_conn in env_vars but in mode FIXED
            mode = DatabaseOdooConfigSection.MaxConnMode.AUTO
        if mode == DatabaseOdooConfigSection.MaxConnMode.AUTO and not self.max_conn:
            self.max_conn = min_conn
            # We add some security because sometime worker open 2 or more connecions (Ex :bus.bus)
            self.max_conn = max(self.max_conn, min_conn, 2)

        if self.filter and not self.show:
            self.show = True
        if self.name and not self.show:
            self.filter = self.name
        if self.name and self.show and not self.filter:
            self.filter = self.name + ".*"

    def to_dict(self):
        if not self.enable:
            return {}
        res = OrderedDict()
        res["--db_host"] = self.host
        res["--db_port"] = self.port
        res["--db_user"] = self.user
        res["--db_password"] = self.password
        res["--database"] = self.name
        res["--no-database-list"] = not self.show
        res["--db_maxconn"] = self.max_conn
        res["--db-filter"] = self.filter
        return res


class HttpOdooConfigSection(OdooConfigSection):
    def __init__(self, odoo_config_maker, env_vars):
        super(HttpOdooConfigSection, self).__init__(odoo_config_maker, env_vars)
        self.enable = self.is_true(env_vars.get("HTTP_ENABLE", "True"))
        self.interface = None
        self.port = None
        self.longpolling_port = None

        if self.enable:
            self.interface = env_vars.get("HTTP_INTERFACE") or "0.0.0.0"
            self.port = self.to_int(env_vars.get("HTTP_PORT")) or 8080
            self.longpolling_port = self.to_int(env_vars.get("LONGPOLLING_PORT")) or 4040

    def to_dict(self):
        key_http = "http" if self.config_maker.odoo_version > 10 else "xmlrpc"
        if not self.enable:
            return {
                "--no-%s" % key_http: not self.enable,
            }
        res = OrderedDict()
        res["--%s-interface" % key_http] = self.interface
        res["--%s-port" % key_http] = self.port
        res["--longpolling-port"] = self.longpolling_port
        return res


class ServerWideModuleConfigSection(OdooConfigSection):
    def __init__(self, odoo_config_maker, env_vars):
        super(ServerWideModuleConfigSection, self).__init__(odoo_config_maker, env_vars)
        str_server_wide_modules = env_vars.get("SERVER_WIDE_MODULES")
        self.server_wide_modules = str_server_wide_modules and str_server_wide_modules.split(",") or ["base", "web"]
        self.queue_job_module_name = None
        if env_vars.get("QUEUE_JOB_ENABLE"):
            self.queue_job_module_name = "queue_job"
            if odoo_config_maker.odoo_version < 10:
                self.queue_job_module_name = "connector"
            self.server_wide_modules.append(self.queue_job_module_name)

        print(
            "enable odoo_filestore_s3 ?",
            self.is_true(env_vars.get("S3_FILESTORE_ENABLE")),
            env_vars.get("S3_FILESTORE_ENABLE"),
        )
        print(
            "enable odoo_session_redis ? ",
            self.is_true(env_vars.get("REDIS_SESSION_ENABLE")),
            env_vars.get("REDIS_SESSION_ENABLE"),
        )
        if self.is_true(env_vars.get("S3_FILESTORE_ENABLE")):
            self.server_wide_modules.append("odoo_filestore_s3")
        if self.is_true(env_vars.get("REDIS_SESSION_ENABLE")):
            self.server_wide_modules.append("odoo_session_redis")

    def remove_queue_job(self):
        if self.queue_job_module_name:
            self.server_wide_modules.remove(self.queue_job_module_name)

    def to_dict(self):
        return {
            "--load": self.server_wide_modules,
        }


class OtherSection(OdooConfigSection):
    def __init__(self, odoo_config_maker, env_vars):
        super(OtherSection, self).__init__(odoo_config_maker, env_vars)
        self.unaccent = self.is_true(env_vars.get("UNACCENT", True))
        self.test_enable = self.is_true(env_vars.get("TEST_ENABLE"))
        self.without_demo = self.is_true(env_vars.get("WITHOUT_DEMO", "all"))

    def to_dict(self):
        if not self.enable:
            return {"--unaccent": self.unaccent}
        res = OrderedDict()
        res["--unaccent"] = self.unaccent
        res["--test-enable"] = self.test_enable
        res["--without-demo"] = self.without_demo
        return res


class LoggerSection(OdooConfigSection):
    def __init__(self, odoo_config_maker, env_vars):
        super(LoggerSection, self).__init__(odoo_config_maker, env_vars)
        self.logfile = env_vars.get("LOGFILE")
        self.log_handler = env_vars.get("LOG_HANDLER")
        self.log_request = self.is_true(env_vars.get("LOG_REQUEST"))
        self.log_response = self.is_true(env_vars.get("LOG_RESPONSE"))
        self.log_web = self.is_true(env_vars.get("LOG_WEB"))
        self.log_sql = self.is_true(env_vars.get("LOG_SQL"))
        self.log_db = self.is_true(env_vars.get("LOG_DB"))
        self.log_db_level = env_vars.get("LOG_DB_LEVEL")
        self.log_level = env_vars.get("LOG_LEVEL")

    def to_dict(self):
        if not self.enable:
            return {}
        res = OrderedDict()
        res["--logfile"] = self.logfile
        res["--log-handler"] = self.log_handler
        res["--log-request"] = self.log_request
        res["--log-response"] = self.log_response
        res["--log-web"] = self.log_web
        res["--log-sql"] = self.log_sql
        res["--log-db"] = self.log_db
        res["--log-db-level"] = self.log_db_level
        res["--log-level"] = self.log_level
        return res


class UpdateInstallSection(OdooConfigSection):
    def __init__(self, odoo_config_maker, env_vars):
        super(UpdateInstallSection, self).__init__(odoo_config_maker, env_vars)
        self.update = [u.strip() for u in env_vars.get("UPDATE", "").split(",")]
        self.install = [i.strip() for i in env_vars.get("INSTALL", "").split(",")]
        self.stop_after_init = self.is_true(env_vars.get("STOP_AFTER_INIT"))
        self.save_config_file = self.is_true(env_vars.get("SAVE_CONFIG_FILE"))
        self.force_stop_after_init = False
        self.force_save_config_file = False

    def to_dict(self):
        default_res = OrderedDict()
        if self.force_save_config_file:
            default_res["--save"] = self.force_save_config_file
        if self.force_stop_after_init:
            default_res["--stop-after-init"] = self.force_stop_after_init
        if not self.enable:
            return default_res
        res = OrderedDict()
        res["--update"] = ",".join(self.update)
        res["--init"] = ",".join(self.install)
        res["--stop-after-init"] = self.stop_after_init
        res["--save"] = self.save_config_file
        res.update(default_res)
        return res


class AddonsPathConfigSection(OdooConfigSection):
    def __init__(self, odoo_config_maker, env_vars):
        super(AddonsPathConfigSection, self).__init__(odoo_config_maker, env_vars)
        registry = addons_installer.AddonsRegistry()
        result = registry.parse_env(env_vars=env_vars)
        self.addons_path = [r.addons_path for r in result]

    def to_dict(self):
        if not self.enable:
            return {}
        return {
            "--addons-path": self.addons_path,
        }



from __future__ import unicode_literals

import logging
import os
import sys

from . import laucher  # noqa
from . import api, config_section, mapper  # noqa
from .odoo_config_maker import OdooConfig, OdooEnvConverter  # noqa

_logger = logging.getLogger(__name__)
# _logger.setLevel(logging.DEBUG)


def main():
    env_vars = dict(os.environ)
    _logger.info("create config")
    odoo_path = os.getenv("ODOO_PATH")
    server_path = os.getenv("NDP_SERVER_PATH")
    odoo_rc = os.getenv("ODOO_RC")
    launcher = laucher.Launcher(sys.argv[1:], odoo_path=odoo_path, odoo_rc=odoo_rc, server_path=server_path)
    launcher.init_addons(env_vars)
    return_code = launcher.launch_config_file(env_vars).wait()
    if return_code:
        sys.exit(return_code)
    if os.getenv("UPDATE") or os.getenv("INSTALL"):
        _logger.info("Update or init detected")
        maintenance_server_proc = launcher.launch_maintenance_server()
        return_code = launcher.launch_update(env_vars).wait()
        maintenance_server_proc.kill()
        if return_code:
            sys.exit(return_code)
    _logger.info("#############################################")
    _logger.info("Run Odoo")
    sys.exit(launcher.launch({}).wait())


if __name__ == "__main__":
    main()
